package sra.isr.image.database;

import java.sql.*;
import java.util.*;

/**
 * IsrConnectionPool class
 * 
 * 	@author    Mitsuhiro Asada
 * 	@created   2000/12/27 (by Mitsuhiro Asada)
 * 	@updated   N/A
 * 	@version   Isr023b
 * 	@J2SDK     1.2.2 or higher
 * 	@copyright 2000-2002 SRA (Software Research Associates, Inc.)
 * 	@copyright 2002 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * 	 All rights reserved.
 * 
 * 	 Redistribution and use in source and binary forms, with or without
 * 	 modification, are permitted provided that the following conditions
 * 	 are met:
 * 	
 * 	 1. Redistributions of source code must retain the above copyright
 * 	    notice, this list of conditions and the following disclaimer. 
 * 	
 * 	 2. Redistributions in binary form must reproduce the above copyright
 * 	    notice, this list of conditions and the following disclaimer in
 * 	    the documentation and/or other materials provided with the
 * 	    distribution. 
 * 	
 * 	 3. The name of the author may not be used to endorse or promote
 * 	    products derived from this software without specific prior
 * 	    written permission. 
 * 	
 * 	 THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * 	 IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * 	 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * 	 ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * 	 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * 	 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * 	 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * 	 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * 	 IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * 	 OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * 	 IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 	
 * 	 Java and all Java-based trademarks and logos are trademarks or
 * 	 registered trademarks of Sun Microsystems, Inc. in the U.S. or
 * 	 other countries, and are used under license. 
 * 
 * $Id: IsrConnectionPool.java,v 4.2 2002/07/11 09:25:03 m-asada Exp $
 */
public class IsrConnectionPool {
	protected static IsrConnectionPool SoleInstance;

	// database parameters.
	private String url;
	private String user;
	private String password;
	private String jdbcDriver;
	private String batchSize;

	// connection pooling.
	protected int minConnectionSize;
	protected int maxConnectionSize;
	protected Stack connectionToken;
	protected Stack connections;
	/**
	 * The default constructor.
	 */
	protected IsrConnectionPool() {
	}
	/**
	 * Create a new connection.
	 *
	 * @return java.sql.Connection
	 * @exception java.sql.SQLException 
	 */
	protected Connection createNewConnection() throws SQLException {
		Properties info = new Properties();
		info.put("user", this.getDatabaseUser());
		info.put("password", this.getDatabasePassword());
		info.put("defaultBatchValue", this.getJdbcBatchSize());
		//
		connectionToken.push(new Object());
		return DriverManager.getConnection(this.getDatabaseUrl(), info);
	}
	/**
	 * Close the database connections.
	 * 
	 * @exception java.lang.Throwable
	 */
	protected void finalize() throws Throwable {
		while (this.getConnections().empty() == false) {
			Connection connection = (Connection) this.getConnections().pop();
			connection.close();
			connection = null;
		}
		super.finalize();
	}
	/**
	 * Answer the connection from connection pool.
	 * 
	 * @return java.sql.Connection
	 * @exception java.sql.SQLException 
	 */
	public synchronized Connection getConnection() throws SQLException {
		if (this.getConnections().empty()) {
			while (connectionToken.empty()) {
				Thread.yield();
			}
			return this.createNewConnection();
		}
		return (Connection) this.getConnections().pop();
	}
	/**
	 * Answer the connections.
	 *
	 * @return java.util.Stack
	 * @exception java.sql.SQLException 
	 */
	protected Stack getConnections() throws SQLException {
		if (connections == null) {
			this.initializeConnections();
		}
		return connections;
	}
	/**
	 * Answer the password of the database user.
	 *
	 * @return java.lang.String
	 */
	protected String getDatabasePassword() {
		return password;
	}
	/**
	 * Answer the url of the database.
	 * 
	 * @return java.lang.String
	 */
	protected String getDatabaseUrl() {
		return url;
	}
	/**
	 * Answer the user name of the database.
	 *
	 * @return java.lang.String
	 */
	protected String getDatabaseUser() {
		return user;
	}
	/**
	 * Answer an instance of connection pool.
	 *
	 * @param properties java.util.Properties
	 * @return sra.isr.database.IsrConnectionPool
	 * @exception java.sql.SQLException 
	 */
	public static synchronized IsrConnectionPool GetInstance(Properties properties) throws SQLException {
		if (SoleInstance == null) {
			SoleInstance = new IsrConnectionPool();
			SoleInstance.setDatabaseUrl(properties.getProperty("database.url"));
			SoleInstance.setDatabaseUser(properties.getProperty("database.user"));
			SoleInstance.setDatabasePassword(properties.getProperty("database.password"));
			SoleInstance.setMinConnectionSize(Integer.parseInt(properties.getProperty("database.minConnectionSize")));
			SoleInstance.setMaxConnectionSize(Integer.parseInt(properties.getProperty("database.maxConnectionSize")));
			SoleInstance.setJdbcDriver(properties.getProperty("database.jdbc.driver"));
			SoleInstance.setJdbcBatchSize(properties.getProperty("database.jdbc.batchSize"));
			SoleInstance.initializeConnections();
		}
		return SoleInstance;
	}
	/**
	 * Answer the batch size of the database.
	 *
	 * @return java.lang.String
	 */
	protected String getJdbcBatchSize() {
		if (batchSize == null) {
			batchSize = "20";
		}
		return batchSize;
	}
	/**
	 * Answer the name of the JDBC driver.
	 *
	 * @return java.lang.String
	 */
	protected String getJdbcDriver() {
		return jdbcDriver;
	}
	/**
	 * Answer the maximum connection size of the database.
	 *
	 * @return int
	 */
	protected int getMaxConnectionSize() {
		if (maxConnectionSize < 1) {
			maxConnectionSize = 1;
		}
		if (maxConnectionSize <= this.getMinConnectionSize()) {
			maxConnectionSize = this.getMinConnectionSize();
		}
		return maxConnectionSize;
	}
	/**
	 * Answer the minimum connection size of the database.
	 *
	 * @return int
	 */
	protected int getMinConnectionSize() {
		if (minConnectionSize < 0) {
			minConnectionSize = 0;
		}
		return minConnectionSize;
	}
	/**
	 * Initialize the connection pool.
	 *
	 * @exception java.sql.SQLException 
	 */
	protected void initializeConnections() throws SQLException {
		connections = new Stack();
		connectionToken = new Stack();
		for (int i =0; i < this.getMaxConnectionSize(); i++) {
			connectionToken.push(new Object());
		}
		//
		try {
			Class.forName(this.getJdbcDriver());
			for (int i = 0; i < this.getMinConnectionSize(); i++) {
				Connection connection = this.createNewConnection();
				if (connection != null) {
					connections.push(connection);
				}
			}
		} catch (ClassNotFoundException e) {
			System.out.println("JDBC Drvier class not found: " + this.getJdbcDriver());
			e.printStackTrace();
		}
	}
	/**
	 * Release and Pool connection.
	 * 
	 * @param connection java.sql.Connection
	 * @exception java.sql.SQLException 
	 */
	public void releaseConnection(Connection connection) throws SQLException {
		if (connection == null) {
			return;
		}
		if (connection.getAutoCommit() == false) {
			connection.rollback();
		}
		if (connection.isClosed() == true) {
			connection = null;
			return;
		}
		if (this.getConnections().size() < this.getMinConnectionSize()) {
			this.getConnections().push(connection);
		} else {
			connection.close();
			connection = null;
			connectionToken.push(new Object());
		}
	}
	/**
	 * Set the password of the database user.
	 *
	 * @param newPassword java.lang.String
	 */
	protected void setDatabasePassword(String newPassword) {
		password = newPassword;
	}
	/**
	 * Set the url of the database.
	 *
	 * @param newUrl java.lang.String
	 */
	protected void setDatabaseUrl(String newUrl) {
		url = newUrl;
	}
	/**
	 * Set the user of the database.
	 *
	 * @param newUser java.lang.String
	 */
	protected void setDatabaseUser(String newUser) {
		user = newUser;
	}
	/**
	 * Set the batch size of the database.
	 *
	 * @param newBatchSize java.lang.String
	 */
	protected void setJdbcBatchSize(String newBatchSize) {
		batchSize = newBatchSize;
	}
	/**
	 * Set the batch size of the database.
	 *
	 * @param newDriver java.lang.String
	 */
	protected void setJdbcDriver(String newDriver) {
		jdbcDriver = newDriver;
	}
	/**
	 * Set the maxmum connection size of the database.
	 *
	 * @param newSize int
	 */
	protected void setMaxConnectionSize(int newSize) {
		maxConnectionSize = newSize;
	}
	/**
	 * Set the minimum connection size of the database.
	 *
	 * @param newSize int
	 */
	protected void setMinConnectionSize(int newSize) {
		minConnectionSize = newSize;
	}
}
