package sra.isr.image.database;

import java.awt.*;
import java.io.*;
import java.sql.*;
import java.util.*;
import sra.smalltalk.*;
import sra.isr.geometry.basic.*;
import sra.isr.image.objects.*;

/**
 * IsrImageDatabase class
 * 
 * 	@author    Mitsuhiro Asada
 * 	@created   2000/12/28 (by Mitsuhiro Asada)
 * 	@updated   2001/07/10 (by Mitsuhiro Asada)
 * 	@version   Isr023b
 * 	@J2SDK     1.2.2 or higher
 * 	@copyright 2000-2002 SRA (Software Research Associates, Inc.)
 * 	@copyright 2002 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * 	 All rights reserved.
 * 
 * 	 Redistribution and use in source and binary forms, with or without
 * 	 modification, are permitted provided that the following conditions
 * 	 are met:
 * 	
 * 	 1. Redistributions of source code must retain the above copyright
 * 	    notice, this list of conditions and the following disclaimer. 
 * 	
 * 	 2. Redistributions in binary form must reproduce the above copyright
 * 	    notice, this list of conditions and the following disclaimer in
 * 	    the documentation and/or other materials provided with the
 * 	    distribution. 
 * 	
 * 	 3. The name of the author may not be used to endorse or promote
 * 	    products derived from this software without specific prior
 * 	    written permission. 
 * 	
 * 	 THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * 	 IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * 	 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * 	 ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * 	 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * 	 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * 	 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * 	 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * 	 IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * 	 OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * 	 IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 	
 * 	 Java and all Java-based trademarks and logos are trademarks or
 * 	 registered trademarks of Sun Microsystems, Inc. in the U.S. or
 * 	 other countries, and are used under license. 
 * 
 * $Id: IsrImageDatabase.java,v 4.2 2002/07/11 09:25:03 m-asada Exp $
 */
public abstract class IsrImageDatabase {
	protected Properties properties = null;
	/**
	 * The default constructor.
	 */
	protected IsrImageDatabase() {
	}
	/**
	 * Create the condition string of average gray color.
	 *
	 * @return java.lang.String
	 * @param luminance double
	 */
	protected String conditionOfAverageGray(double luminance) {
		String condition = null;
		if (0.0d <= luminance && luminance <= 100d) {
			condition = "CAST(ABS(luminance - " + (int) (luminance * 255 / 100) + ") AS FLOAT)";
		}
		return condition;
	}
	/**
	 * Create the condition string of center spectra.
	 *
	 * @return java.lang.String
	 * @param centerSpectra sra.isr.geometry.basic.Isr2dPoint
	 */
	protected String conditionOfCenterSpectra(Isr2dPoint centerSpectra) {
		String condition = null;
		if (centerSpectra != null) {
			Vector collection = new Vector(2);
			collection.addElement("CAST(ABS(center_x - " + (int) (centerSpectra.x() * 255) + ") AS FLOAT)");
			collection.addElement("CAST(ABS(center_y - " + (int) (centerSpectra.y() * 255) + ") AS FLOAT)");
			condition = this.createConditionSql(collection);
		}
		return condition;
	}
	/**
	 * Create the condition string of icon extent.
	 *
	 * @return java.lang.String
	 * @param iconArea sra.isr.geometry.basic.Isr2dPoint
	 */
	protected String conditionOfIconArea(Isr2dPoint iconArea) {
		String condition = null;
		if (iconArea != null) {
			condition = "CAST(ABS(width_ratio - " + (int) (iconArea.x() * 255 / (iconArea.x() + iconArea.y())) + ") AS FLOAT)";
		}
		return condition;
	}
	/**
	 * Create the condition string of nine colors.
	 *
	 * @return java.lang.String
	 * @param nineColors java.awt.Color[]
	 */
	protected String conditionOfNineColors(Color[] nineColors) {
		String condition = null;
		if (nineColors != null) {
			Vector collection = new Vector();
			for (int i = 0; i < nineColors.length; i++) {
				if (nineColors[i] != null) {
					collection.addElement("CAST(ABS(nc" + (i + 1) + "_r - " + nineColors[i].getRed() + ") AS FLOAT)");
					collection.addElement("CAST(ABS(nc" + (i + 1) + "_g - " + nineColors[i].getGreen() + ") AS FLOAT)");
					collection.addElement("CAST(ABS(nc" + (i + 1) + "_b - " + nineColors[i].getBlue() + ") AS FLOAT)");
				}
			}
			condition = this.createConditionSql(collection);
		}
		return condition;
	}
	/**
	 * Create the condition string of typical color.
	 *
	 * @return java.lang.String
	 * @param typicalColor java.awt.Color
	 */
	protected String conditionOfTypicalColor(Color typicalColor) {
		String condition = null;
		if (typicalColor != null) {
			Vector collection = new Vector(3);
			collection.addElement("CAST(ABS(tc_r - " + typicalColor.getRed() + ") AS FLOAT)");
			collection.addElement("CAST(ABS(tc_g - " + typicalColor.getGreen() + ") AS FLOAT)");
			collection.addElement("CAST(ABS(tc_b - " + typicalColor.getBlue() + ") AS FLOAT)");
			condition = this.createConditionSql(collection);
		}
		return condition;
	}
	/**
	 * Create the condition strings of all parameters.
	 *
	 * @return java.util.Hashtable[] or null
	 * @param typicalColor java.awt.Color
	 * @param double luminance
	 * @param nineColors java.awt.Color[]
	 * @param iconArea sra.isr.geometry.basic.Isr2dPoint
	 * @param centerSpectra sra.isr.geometry.basic.Isr2dPoint
	 */
	protected Vector createAllConditions(Color typicalColor, double luminance, Color[] nineColors, Isr2dPoint iconArea, Isr2dPoint centerSpectra) {
		Vector conditions = new Vector();
		// typical color
		String conditionOfTypicalColor = this.conditionOfTypicalColor(typicalColor);
		if (conditionOfTypicalColor != null) {
			conditions.addElement(conditionOfTypicalColor);
		}
		// luminance
		String conditionOfAverageGray = this.conditionOfAverageGray(luminance);
		if (conditionOfAverageGray != null) {
			conditions.addElement(conditionOfAverageGray);
		}
		// nine colors
		String conditionOfNineColors = this.conditionOfNineColors(nineColors);
		if (conditionOfNineColors != null) {
			conditions.addElement(conditionOfNineColors);
		}
		// icon area
		String conditionOfIconArea = this.conditionOfIconArea(iconArea);
		if (conditionOfIconArea != null) {
			conditions.addElement(conditionOfIconArea);
		}
		// center spectra
		String conditionOfCenterSpectra = this.conditionOfCenterSpectra(centerSpectra);
		if (conditionOfCenterSpectra != null) {
			conditions.addElement(conditionOfCenterSpectra);
		}
		return conditions;
	}
	/**
	 * Answer the sql string of connditions.
	 *
	 * @return java.lang.String
	 * @param conditions java.lang.String[]
	 */
	protected String createConditionSql(String[] conditions) {
		String result = null;
		StringWriter sw = new StringWriter();
		PrintWriter pw = new PrintWriter(sw);
		try {
			pw.write("(");
			//
			for (int i = 0; i < conditions.length - 1; i++) {
				pw.write(conditions[i]);
				pw.write(" + ");
			}
			pw.write(conditions[conditions.length - 1]);
			//
			pw.write(") / " + conditions.length);
			result = sw.toString();
		} finally {
			pw.flush();
			pw.close();
			pw = null;
		}
		return result;
	}
	/**
	 * Answer the sql string of connditions.
	 *
	 * @return java.lang.String
	 * @param conditions java.util.Vector
	 */
	protected String createConditionSql(Vector conditions) {
		String[] strings = new String[conditions.size()];
		conditions.copyInto(strings);
		return this.createConditionSql(strings);
	}
	/**
	 * Get the connection from connection pool.
	 * 
	 * @return java.sql.Connection
	 * @exception java.sql.SQLException 
	 */
	protected Connection getConnection() throws SQLException {
		return this.getConnectionPool().getConnection();
	}
	/**
	 * Get the instance of connection pool.
	 *
	 * @return sra.isr.image.database.IsrConnectionPool
	 * @exception java.sql.SQLException 
	 */
	protected IsrConnectionPool getConnectionPool() throws SQLException {
		if (this.getProperties() != null) {
			return IsrConnectionPool.GetInstance(this.getProperties());
		}
		return null;
	}
	/**
	 * Answer an instance of iamge database.
	 *
	 * @return sra.isr.database.IsrImageDatabase
	 * @param properties java.util.Properties
	 */
	public static IsrImageDatabase GetInstance(Properties properties) {
		IsrImageDatabase database = null;
		try {
			database = (IsrImageDatabase) Class.forName(properties.getProperty("database.classname")).newInstance();
			database.setProperties(properties);
		} catch (Exception e) {
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		return database;
	}
	/**
	 * Get the original image information with primary key on database.
	 *
	 * @return sra.isr.image.objects.IsrImageFileInfo
	 * @param pkey int
	 * @exception java.sql.SQLException
	 */
	public IsrImageFileInfo getOriginalImageInfo(int pkey) throws SQLException {
		IsrImageFileInfo imageInfo = new IsrImageFileInfo();
		Connection connection = null;

		try {
			connection = this.getConnection();
			connection.setTransactionIsolation(Connection.TRANSACTION_SERIALIZABLE);
			connection.setAutoCommit(false);
			//
			imageInfo = this.getOriginalImageInfo(connection, pkey);
			//
			connection.commit();
		} catch (SQLException e) {
			connection.rollback();
			throw e;
		} finally {
			if (connection != null) {
				this.releaseConnection(connection);
				connection = null;
			}
		}

		return imageInfo;
	}
	/**
	 * Get the original image information with connection and primary key on database.
	 *
	 * @return sra.isr.image.objects.IsrImageFileInfo
	 * @param connection java.sql.Connection
	 * @param pkey int
	 * @exception java.sql.SQLException
	 */
	abstract protected IsrImageFileInfo getOriginalImageInfo(Connection connection, int pkey) throws SQLException;
	/**
	 * Get the properties of database.
	 * 
	 * @return java.util.Properties
	 */
	protected Properties getProperties() {
		return properties;
	}
	/**
	 * Get the thumbnail image information with primary key on database.
	 *
	 * @return sra.isr.image.objects.IsrImageFileInfo
	 * @param pkey int
	 * @exception java.sql.SQLException
	 */
	public IsrImageFileInfo getThumbnailImageInfo(int pkey) throws SQLException {
		IsrImageFileInfo imageInfo = null;
		Connection connection = null;
		
		try {
			connection = this.getConnection();
			connection.setTransactionIsolation(Connection.TRANSACTION_SERIALIZABLE);
			connection.setAutoCommit(false);
			//
			imageInfo = this.getThumbnailImageInfo(connection, pkey);
			//
			connection.commit();
		} catch (SQLException e) {
			connection.rollback();
			throw e;
		} finally {
			if (connection != null) {
				this.releaseConnection(connection);
				connection = null;
			}
		}
		
		return imageInfo;
	}
	/**
	 * Get the thumbnail image information with connection and primary key on database.
	 *
	 * @return sra.isr.image.objects.IsrImageFileInfo
	 * @param connection java.sql.Connection
	 * @param pkey int
	 * @exception java.sql.SQLException
	 */
	abstract protected IsrImageFileInfo getThumbnailImageInfo(Connection connection, int pkey) throws SQLException;
	/**
	 * Get the next value of image searcher primary key.
	 *
	 * @return int
	 * @param connection java.sql.Connection
	 * @exception java.sql.SQLException
	 */
	abstract protected int nextImageSearcherPkey(Connection connection) throws SQLException;
	/**
	 * Release the connection.
	 * 
	 * @param connection java.sql.Connection
	 * @exception java.sql.SQLException 
	 */
	protected void releaseConnection(Connection connection) throws SQLException {
		this.getConnectionPool().releaseConnection(connection);
	}
	/**
	 * Search images with parameters.
	 *
	 * @return java.util.Hashtable[] or null
	 * @param typicalColor java.awt.Color
	 * @param double luminance
	 * @param nineColors java.awt.Color[]
	 * @param iconArea sra.isr.geometry.basic.Isr2dPoint
	 * @param centerSpectra sra.isr.geometry.basic.Isr2dPoint
	 * @param double cutOff
	 * @exception java.sql.SQLException
	 */
	public Hashtable[] search(Color typicalColor, double luminance, Color[] nineColors, Isr2dPoint iconArea, Isr2dPoint centerSpectra, double cutOff) throws SQLException {
		Connection connection = null;
		Hashtable[] ids = null;
		Vector conditions = this.createAllConditions(typicalColor, luminance, nineColors, iconArea, centerSpectra);

		try {
			connection = this.getConnection();
			connection.setTransactionIsolation(Connection.TRANSACTION_SERIALIZABLE);
			connection.setAutoCommit(false);
			//
			if (conditions.size() > 0) {
				String sqlConditions = "(CAST(" + this.createConditionSql(conditions) + " AS FLOAT) / 255.0)";
				ids = this.searchWithConditions(connection, sqlConditions, cutOff);
			} else {
				ids = this.searchNullConditions(connection);
			}
			//
			connection.commit();
		} catch (SQLException e) {
			connection.rollback();
			throw e;
		} finally {
			if (connection != null) {
				this.releaseConnection(connection);
				connection = null;
			}
		}

		return ids;
	}
	/**
	 * Search all images.
	 *
	 * @return java.util.Hashtable[]
	 * @param connection java.sql.Connection
	 * @exception java.sql.SQLException
	 */
	protected Hashtable[] searchNullConditions(Connection connection) throws SQLException {
		String sql = null;
		PreparedStatement pst = null;
		ResultSet rs = null;
		Vector results = new Vector();

		try {
			sql = "SELECT pkey, icon_area_x, icon_area_y FROM image_searcher ORDER BY pkey";
			pst = connection.prepareStatement(sql);
			pst.clearParameters();
			rs = pst.executeQuery();
			if (rs != null) {
				while (rs.next()) {
					Hashtable answer = new Hashtable();
					answer.put("pkey", new Integer(rs.getInt(1)));
					answer.put("value", new Float(1.0f));
					answer.put("iconArea", new Dimension(rs.getInt(2), rs.getInt(3)));
					results.addElement(answer);
				}
				rs.close();
				rs = null;
			}
			pst.close();
			pst = null;
		} finally {
			if (rs != null) {
				rs.close();
				rs = null;
			}
			if (pst != null) {
				pst.close();
				pst = null;
			}
		}

		Hashtable[] ids = new Hashtable[results.size()];
		results.copyInto(ids);
		return ids;
	}
	/**
	 * Search images with sql conditions.
	 *
	 * @return java.util.Hashtable[]
	 * @param connection java.sql.Connection
	 * @param sqlConditions java.lang.String
	 * @param double cutOff
	 * @exception java.sql.SQLException
	 */
	abstract protected Hashtable[] searchWithConditions(Connection connection, String sqlConditions, double cutOff) throws SQLException;
	/**
	 * Set the properties of database.
	 * 
	 * @param properties java.util.Properties
	 */
	protected void setProperties(Properties newProperties) {
		properties = newProperties;
	}
	/**
	 * Store an image object on database with pkey.
	 *
	 * @param connection java.sql.Connection
	 * @param pkey int
	 * @param imageObject sra.isr.image.objects.IsrImageObject
	 * @param originalImage sra.isr.image.objects.IsrImageFileInfo
	 * @param thumbnailImage sra.isr.image.objects.IsrImageFileInfo
	 * @exception java.sql.SQLException
	 */
	abstract protected void storeImage(Connection connection, int pkey, IsrImageObject imageObject, IsrImageFileInfo originalImage, IsrImageFileInfo thumbnailImage) throws SQLException;
	/**
	 * Store an image object on database.
	 *
	 * @return int
	 * @param imageObject sra.isr.image.objects.IsrImageObject
	 * @param originalImage sra.isr.image.objects.IsrImageFileInfo
	 * @param thumbnailImage sra.isr.image.objects.IsrImageFileInfo
	 * @exception java.sql.SQLException
	 */
	public int storeImage(IsrImageObject imageObject, IsrImageFileInfo originalImage, IsrImageFileInfo thumbnailImage) throws SQLException {
		Connection connection = null;
		int pkey = -1;

		try {
			connection = this.getConnection();
			connection.setTransactionIsolation(Connection.TRANSACTION_READ_COMMITTED);
			connection.setAutoCommit(false);
			//
			pkey = this.nextImageSearcherPkey(connection);
			this.storeImage(connection, pkey, imageObject, originalImage, thumbnailImage);
			//
			connection.commit();
		} catch (SQLException e) {
			connection.rollback();
			throw e;
		} finally {
			if (connection != null) {
				this.releaseConnection(connection);
				connection = null;
			}
		}
		return pkey;
	}
}
