package sra.isr.image.objects;

import java.io.*;
import java.util.*;
import sra.smalltalk.*;
import sra.isr.image.streams.*;

/**
 * IsrImageFileInfo class
 * 
 * 	@author    Mitsuhiro Asada
 * 	@created   2001/06/16 (by Mitsuhiro Asada)
 * 	@updated   N/A
 * 	@version   Isr023b
 * 	@J2SDK     1.2.2 or higher
 * 	@copyright 2000-2002 SRA (Software Research Associates, Inc.)
 * 	@copyright 2002 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * 	 All rights reserved.
 * 
 * 	 Redistribution and use in source and binary forms, with or without
 * 	 modification, are permitted provided that the following conditions
 * 	 are met:
 * 	
 * 	 1. Redistributions of source code must retain the above copyright
 * 	    notice, this list of conditions and the following disclaimer. 
 * 	
 * 	 2. Redistributions in binary form must reproduce the above copyright
 * 	    notice, this list of conditions and the following disclaimer in
 * 	    the documentation and/or other materials provided with the
 * 	    distribution. 
 * 	
 * 	 3. The name of the author may not be used to endorse or promote
 * 	    products derived from this software without specific prior
 * 	    written permission. 
 * 	
 * 	 THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * 	 IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * 	 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * 	 ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * 	 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * 	 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * 	 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * 	 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * 	 IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * 	 OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * 	 IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 	
 * 	 Java and all Java-based trademarks and logos are trademarks or
 * 	 registered trademarks of Sun Microsystems, Inc. in the U.S. or
 * 	 other countries, and are used under license. 
 * 
 * $Id: IsrImageFileInfo.java,v 4.3 2002/07/11 09:25:03 m-asada Exp $
 */
public class IsrImageFileInfo {
	protected static Hashtable ImageContentTypeTable;
	protected String name;
	protected String filename;
	protected String contentType;
	protected byte[] bytes;
	//
	static {
		ImageContentTypeTable = InitializeImageContentTypeTable();
	}
	/**
	 * The default constructor. 
	 */
	public IsrImageFileInfo() {
	}
	/**
	 * The default constructor with content-type and image. 
	 *
	 * @param name java.lang.String
	 * @param filename java.lang.String
	 * @param contentType java.lang.String
	 * @param image byte[]
	 */
	public IsrImageFileInfo(String contentType, byte[] image) {
		this(null, null, contentType, image);
	}
	/**
	 * The default constructor with filename, content-type and image. 
	 *
	 * @param name java.lang.String
	 * @param filename java.lang.String
	 * @param contentType java.lang.String
	 * @param image byte[]
	 */
	public IsrImageFileInfo(String filename, String contentType, byte[] image) {
		this(null, filename, contentType, image);
	}
	/**
	 * The default constructor with name, filename, content-type and image. 
	 *
	 * @param name java.lang.String
	 * @param filename java.lang.String
	 * @param contentType java.lang.String
	 * @param bytes byte[]
	 */
	public IsrImageFileInfo(String name, String filename, String contentType, byte[] bytes) {
		this();
		this.setName(name);
		this.setFilename(filename);
		this.setContentType(contentType);
		this.setBytes(bytes);
	}
	/**
	 * Answer the image object from this image.
	 *
	 * @return sra.isr.image.objects.IsrImageObject
	 * @exception java.io.IOException
	 */
	public IsrImageObject asImageObject() throws IOException {
		String className = this.imageStreamClass(this.getContentType());
		if (className == null) {
			throw new IOException("The content type '" + this.getContentType() + "' is not supported.");
		}
		IsrImageObject imageObject = null;
		IsrImageStream imageStream = null;
		try {
			imageStream = (IsrImageStream) StObject._PerformWith(Class.forName(className), "On_", new ByteArrayInputStream(this.getBytes()));
			imageObject = IsrImageAuxiliary.OriginalImage_(imageStream.nextImage());
		} catch (ClassNotFoundException e) {
			throw new IOException("The class of '" + className + "' is not found.");
		} catch (NoSuchMethodException e) {
			throw new IOException("The method of 'On_' is not such at '" + className + "'.");
		} finally {
			if (imageStream != null) {
				imageStream.close();
				imageStream = null;
			}
		}

		imageObject.iconImage();
		imageObject.iconArea();
		imageObject.nineColors();
		imageObject.averageColor();
		imageObject.typicalColor();
		imageObject.centerSpectra();
		// imageObject.xSpectrum();
		// imageObject.ySpectrum();
		// imageObject.renderingImage();
		return imageObject;
	}
	/**
	 * Get the bytes of this image
	 *
	 * @return byte[]
	 */
	public byte[] getBytes() {
		return bytes;
	}
	/**
	 * Get the content type of this image.
	 *
	 * @return java.lang.String
	 */
	public String getContentType() {
		return contentType;
	}
	/**
	 * Get the filename of this image.
	 *
	 * @return java.lang.String
	 */
	public String getFilename() {
		return filename;
	}
	/**
	 * Get the name of this image.
	 *
	 * @return java.lang.String
	 */
	public String getName() {
		return name;
	}
	/**
	 * Answer the class name of image stream.
	 *
	 * @return java.lang.String
	 * @param contentType java.lang.String
	 */
	protected String imageStreamClass(String contentType) {
		return (String) ImageContentTypeTable.get(contentType);
	}
	/**
	 * Initialize image content type table.
	 *
	 * @return java.util.Hashtable
	 */
	private static Hashtable InitializeImageContentTypeTable() {
		if (ImageContentTypeTable == null) {
			Hashtable table = new Hashtable();
			table.put("image/bmp", "sra.isr.image.streams.IsrBmpImageStream");
			table.put("image/x-bmp", "sra.isr.image.streams.IsrBmpImageStream");
			table.put("image/jpeg", "sra.isr.image.streams.IsrJpegImageStream");
			table.put("image/pjpeg", "sra.isr.image.streams.IsrJpegImageStream");
			table.put("image/png", "sra.isr.image.streams.IsrPngImageStream");
			table.put("image/x-png", "sra.isr.image.streams.IsrPngImageStream");
			ImageContentTypeTable = table;
		}
		return ImageContentTypeTable;
	}
	/**
	 * Set the bytes of this image.
	 *
	 * @param newBytes byte[]
	 */
	public void setBytes(byte[] newBytes) {
		bytes = newBytes;
	}
	/**
	 * Set content type of this image.
	 *
	 * @param newContentType java.lang.String
	 */
	public void setContentType(String newContentType) {
		contentType = newContentType;
	}
	/**
	 * Set filename this image.
	 *
	 * @param newFilename java.lang.String
	 */
	public void setFilename(String newFilename) {
		filename = newFilename;
	}
	/**
	 * Set name this image.
	 *
	 * @param newName java.lang.String
	 */
	public void setName(String newName) {
		name = newName;
	}
}
