package sra.isr.image.streams;

import java.io.*;
import java.util.*;
import sra.smalltalk.*;

/**
 * IsrImageStream class
 * 
 * 	@author    Hirotsugu Kondo
 * 	@created   1999/01/20 (by Hirotsugu Kondo)
 * 	@updated   2000/11/12 (by Mitsuhiro Asada)
 * 	@version   Isr023b
 * 	@J2SDK     1.2.2 or higher
 * 	@copyright 2000-2002 SRA (Software Research Associates, Inc.)
 * 	@copyright 2002 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * 	 All rights reserved.
 * 
 * 	 Redistribution and use in source and binary forms, with or without
 * 	 modification, are permitted provided that the following conditions
 * 	 are met:
 * 	
 * 	 1. Redistributions of source code must retain the above copyright
 * 	    notice, this list of conditions and the following disclaimer. 
 * 	
 * 	 2. Redistributions in binary form must reproduce the above copyright
 * 	    notice, this list of conditions and the following disclaimer in
 * 	    the documentation and/or other materials provided with the
 * 	    distribution. 
 * 	
 * 	 3. The name of the author may not be used to endorse or promote
 * 	    products derived from this software without specific prior
 * 	    written permission. 
 * 	
 * 	 THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * 	 IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * 	 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * 	 ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * 	 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * 	 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * 	 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * 	 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * 	 IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * 	 OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * 	 IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 	
 * 	 Java and all Java-based trademarks and logos are trademarks or
 * 	 registered trademarks of Sun Microsystems, Inc. in the U.S. or
 * 	 other countries, and are used under license. 
 * 
 * $Id: IsrImageStream.java,v 4.2 2002/07/11 09:25:04 m-asada Exp $
 */
public abstract class IsrImageStream extends StObject {
	protected InputStream inStream;
	protected OutputStream outStream;
	protected StImage imageObject;
	protected StValueHolder progressValue = new StValueHolder(0.0f);
	//
	private static Hashtable ImageKindTable;
	static {
		ImageKindTable = _InitializeImageKindTable();
	}
	/**
	 * Initialize image kind table.
	 *
	 * @return java.util.Hashtable
	 */
	private static Hashtable _InitializeImageKindTable() {
		if (ImageKindTable == null) {
			Hashtable aDictionary = new Hashtable();
			aDictionary.put("bmp", "sra.isr.image.streams.IsrBmpImageStream");
			aDictionary.put("jpg", "sra.isr.image.streams.IsrJpegImageStream");
			aDictionary.put("jpeg", "sra.isr.image.streams.IsrJpegImageStream");
			aDictionary.put("png", "sra.isr.image.streams.IsrPngImageStream");
			ImageKindTable = aDictionary;
		}
		return ImageKindTable;
	}
	/**
	 * This image stream close.
	 * 
	 * @exception java.io.IOException
	 */
	public void close() throws IOException {
		if (inStream != null) {
			inStream.close();
			inStream = null;
		}
		if (outStream != null) {
			outStream.close();
			outStream = null;
		}
	}
	/**
	 * Do something with the progress value.
	 *
	 * @param aBlock sra.smalltalk.StBlockClosure
	 */
	public void compute_(StBlockClosure aBlock) {
		this.progressValue().compute_(aBlock);
	}
	/**
	 * This image stream flush.
	 * 
	 * @exception java.io.IOException
	 */
	public void flush() throws IOException {
	    if (outStream != null) {
	        outStream.flush();
	    }
	}
	/**
	 * Answer a Class of image stream.
	 *
	 * @return java.lang.Class
	 * @param aFilename java.io.File
	 */
	public static Class ImageStreamClassForFileName_(File aFilename) {
		return ImageStreamClassForFileName_(aFilename.toString());
	}
	/**
	 * Answer a Class of image stream.
	 *
	 * @return java.lang.Class
	 * @param aFilename java.lang.String
	 */
	public static Class ImageStreamClassForFileName_(String aFilename) {
		String aString = aFilename.substring(aFilename.lastIndexOf('.') + 1).toLowerCase();
		String aSymbol = (String) ImageKindTable.get(aString);
		if (aSymbol == null) {
			return null;
		}
		Class aClass = null;
		try {
			aClass = Class.forName(aSymbol);
		} catch (ClassNotFoundException e) {
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		return aClass;
	}
	/**
	 * Read the image on input stream.
	 *
	 * @return sra.smalltalk.StImage
	 * @exception java.io.IOException
	 */
	public abstract StImage nextImage() throws IOException;
	/**
	 * Write the image on output stream.
	 * 
	 * @param newImage sra.smalltalk.StImage
	 * @exception java.io.IOException
	 */
	public abstract void nextPutImage_(StImage newImage) throws IOException;
	/**
	 * Constractor with input stream.
	 *
	 * @return sra.isr.image.streams.IsrImageStream 
	 * @param stream java.io.InputStream
	 * @exception java.io.IOException
	 */
	public static IsrImageStream On_(InputStream stream) throws IOException {
		throw new IOException("subclassResponsibilitty");
	}
	/**
	 * Set the input stream.
	 * 
	 * @param stream java.io.InputStream
	 * @exception java.io.IOException
	 */
	protected void on_(InputStream stream) throws IOException {
		inStream = stream;
	}
	/**
	 * Constractor with output stream.
	 *
	 * @return sra.isr.image.streams.IsrImageStream 
	 * @param stream java.io.OutputStream
	 * @exception java.io.IOException
	 */
	public static IsrImageStream On_(OutputStream stream) throws IOException {
		throw new IOException("subclassResponsibilitty");
	}
	/**
	 * Set the output stream.
	 * 
	 * @param stream java.io.OutputStream
	 * @exception java.io.IOException
	 */
	protected void on_(OutputStream stream) throws IOException {
		outStream = stream;
	}
	/**
	 * Constractor with image stream and input stream.
	 *
	 * @return sra.isr.image.streams.IsrImageStream 
	 * @param imageStream sra.isr.image.streams.IsrImageStream
	 * @param stream java.io.InputStream
	 * @exception java.io.IOException
	 */
	protected static IsrImageStream On_(IsrImageStream imageStream, InputStream stream) throws IOException {
		if (imageStream == null) {
			throw new IOException("could not load image");
		}
		imageStream.on_(stream);
		return imageStream;
	}
	/**
	 * Constractor with image stream and output stream.
	 *
	 * @return sra.isr.image.streams.IsrImageStream 
	 * @param imageStream sra.isr.image.streams.IsrImageStream
	 * @param stream java.io.OutputStream
	 * @exception java.io.IOException
	 */
	protected static IsrImageStream On_(IsrImageStream imageStream, OutputStream stream) throws IOException {
		if (imageStream == null) {
			throw new IOException("could not save image");
		}
		imageStream.on_(stream);
		return imageStream;
	}
	/**
	 * Set the progress value.
	 *
	 * @param normalizedNumber float
	 */
	public void progress_(float normalizedNumber) {
		if (0 <= normalizedNumber && normalizedNumber <= 1) {
			float truncatedValue = Math.round(normalizedNumber / 0.005) * 0.005f;
			if (this.progressValue()._floatValue() != truncatedValue) {
				this.progressValue().value_(truncatedValue);
			}
		}
	}
	/**
	 * Answer the StValueHolder for the progress value.
	 *
	 * @return sra.smalltalk.StValueHolder
	 */
	protected StValueHolder progressValue() {
		return progressValue;
	}
}
