package sra.smalltalk;

import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;

/**
 * StImage class
 * 
 * 	@author    Hirotsugu Kondo
 * 	@created   1998/11/10 (by Hirotsugu Kondo)
 * 	@updated   2000/04/05 (by Mitsuhiro Asada)
 * 	@updated   2002/04/17 (by nisinaka)
 * 	@version   4.2
 * 	@J2SDK     1.2.2 or higher
 * 	@copyright 1999-2002 SRA (Software Research Associates, Inc.)
 * 	@copyright 2001-2002 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * 	 All rights reserved.
 * 
 * 	 Redistribution and use in source and binary forms, with or without
 * 	 modification, are permitted provided that the following conditions
 * 	 are met:
 * 	
 * 	 1. Redistributions of source code must retain the above copyright
 * 	    notice, this list of conditions and the following disclaimer. 
 * 	
 * 	 2. Redistributions in binary form must reproduce the above copyright
 * 	    notice, this list of conditions and the following disclaimer in
 * 	    the documentation and/or other materials provided with the
 * 	    distribution. 
 * 	
 * 	 3. The name of the author may not be used to endorse or promote
 * 	    products derived from this software without specific prior
 * 	    written permission. 
 * 	
 * 	 THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * 	 IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * 	 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * 	 ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * 	 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * 	 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * 	 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * 	 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * 	 IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * 	 OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * 	 IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 	
 * 	 Java and all Java-based trademarks and logos are trademarks or
 * 	 registered trademarks of Sun Microsystems, Inc. in the U.S. or
 * 	 other countries, and are used under license. 
 * 
 * $Id: StImage.java,v 4.4 2002/07/11 09:18:29 m-asada Exp $
 */
public class StImage extends StObject implements Serializable {
    // Serialize version
    private final static long serialVersionUID = -348945898293926594L;

    transient protected BufferedImage image;
    transient protected int rowByteSize;
    transient protected byte[] bits;

    // Constants
    public final static int WriteZeros = 0;
    public final static int And = 1;
    public final static int Over = 3;
    public final static int Erase = 4;
    public final static int Reverse = 6;
    public final static int Under = 7;
    public final static int ReverseUnder = 13;
    public final static int WriteOnes = 15;
    public final static int Paint = 16;
	/**
	 * Create a new instance of StImage.
	 * 
	 * @param width int
	 * @param height int
	 */
	public StImage(int width, int height) {
	    this(width, height, (int[]) null);

	    int[] pixels = new int[width * height];
	    for (int i = 0; i < pixels.length; i++) {
	        pixels[i] = -1;
	    }
	    this.setPixels(pixels);
	}
    /**
     * Create a new instance of StImage and initialize it.
     *
     * @param width int
     * @param height int
     * @param bits byte[]
     */
    public StImage(int width, int height, byte[] bits) {
        this(width, height, bits, ColorModel.getRGBdefault());
    }
	/**
	 * Create a new instance of StImage and initialize it.
	 *
	 * @param width int
	 * @param height int
	 * @param byteArray byte[]
	 * @param colorModel java.awt.image.ColorModel
	 */
	public StImage(
	    int width,
	    int height,
	    byte[] byteArray,
	    ColorModel colorModel) {

	    image =
	        new BufferedImage(
	            colorModel,
	            colorModel.createCompatibleWritableRaster(width, height),
	            false,
	            null);
	    rowByteSize = (width * this.bitsPerPixel() + 31) >> 5 << 2;

	    if (byteArray != null) {
	        if (rowByteSize * height > byteArray.length) {
	            throw SmalltalkException.Error("Bitmap array has incorrect size.");
	        }

	        // Set the pixels.
	        for (int y = 0; y < height; y++) {
	            for (int x = 0; x < width; x++) {
	                this.setPixel(x, y, this._getPixelFromBits(x, y, byteArray));
	            }
	        }

	        // Set the bit array.
	        bits = byteArray;
	    }
	}
    /**
     * Create a new instance of StImage and initialize it.
     *
     * @param width int
     * @param height int
     * @param pixels int[]
     */
    public StImage(int width, int height, int[] pixels) {
        this(width, height, pixels, ColorModel.getRGBdefault());
    }
	/**
	 * Create a new instance of StImage and initialize it.
	 *
	 * @param width int
	 * @param height int
	 * @param pixels int[]
	 * @param colorModel java.awt.image.ColorModel
	 */
	public StImage(int width, int height, int[] pixels, ColorModel colorModel) {
	    image =
	        new BufferedImage(
	            colorModel,
	            colorModel.createCompatibleWritableRaster(width, height),
	            false,
	            null);
	    rowByteSize = (width * this.bitsPerPixel() + 31) >> 5 << 2;

	    this.setPixels(pixels);
	}
	/**
	 * Create a new instance of StImage and initialize it.
	 *
	 * @param width int
	 * @param height int
	 * @param byteArray sra.smalltalk.StByteArray
	 * @param colorModel java.awt.image.ColorModel
	 */
	public StImage(
	    int width,
	    int height,
	    StByteArray byteArray,
	    ColorModel colorModel) {

	    this(width, height, byteArray._asBytes(), colorModel);
	}
	/**
	 * Create a new instance of StImage and initialize it with the image.
	 *
	 * @param originalImage java.awt.Image
	 */
	public StImage(Image originalImage) {
	    if (originalImage instanceof BufferedImage) {
	        BufferedImage bufferedImage = (BufferedImage) originalImage;
	        image =
	            bufferedImage.getSubimage(
	                0,
	                0,
	                bufferedImage.getWidth(),
	                bufferedImage.getHeight());
	    } else {
	        Component medium = new Canvas();
	        int width = originalImage.getWidth(medium);
	        int height = originalImage.getHeight(medium);
	        int[] pixels = new int[width * height];
	        PixelGrabber grabber =
	            new PixelGrabber(originalImage, 0, 0, width, height, pixels, 0, width);
	        try {
	            grabber.grabPixels();
	        } catch (InterruptedException e) {
	        }
	        image = (new StImage(width, height, pixels)).image();
	        this.setPixels(pixels);
	    }

	    rowByteSize = (this.width() * this.bitsPerPixel() + 31) >> 5 << 2;
	}
    /**
     *
     * @return int
     * @param sourceColor int
     * @param destinationColor int
     * @param combinationRule int
     */
    public int _combinationedColorFrom_to_in_(
        int sourceColor,
        int destinationColor,
        int combinationRule) {
        if (combinationRule == WriteZeros) {
            return 0xFFFFFFFF;
        }
        if (combinationRule == And) {
            return sourceColor & destinationColor;
        }
        if (combinationRule == Over) {
            return sourceColor;
        }
        if (combinationRule == Under) {
            return sourceColor | destinationColor;
        }
        if (combinationRule == WriteOnes) {
            return 0xFF000000;
        }
        return sourceColor;
    }
	/**
	 * Answer image representing my contents using the palette aPalette.
	 * Use ErrorDiffusionImageRender to render the new image.
	 * ErrorDiffusion generate halftones using the Floyd-Steinberg errod diffusion algorithm.
	 * 
	 * @return sra.smalltalk.StImage
	 * @param aPalette java.awt.image.IndexColorModel
	 */
	public StImage _convertToPalette_RenderedByErrorDiffusion(IndexColorModel aPalette) {
	    StImage newImage = (StImage) this.clone();
	    int width = this.width();
	    int height = this.height();
	    Hashtable colorDictionary = new Hashtable();
	    for (int y = 0; y < height; y++) {
	        for (int x = 0; x < width; x++) {
	            Integer pixel = new Integer(newImage.getPixel(x, y));
	            Color currentColor = new Color(pixel.intValue());
	            if (colorDictionary.containsKey(pixel)) {
	                newImage.setPixel(x, y, ((Integer) colorDictionary.get(pixel)).intValue());
	            } else {
	                int fit = StColorValue._MaxDistanceSquared() + 1;
	                int fitRGB = 0xFF000000;
	                for (int j = 0; j < aPalette.getMapSize(); j++) {
	                    int pixelRGB = aPalette.getRGB(j);
	                    Color newColor = new Color(pixelRGB);
	                    Integer d =
	                        StColorValue._DistanceSquaredFrom_ifLessThan_(newColor, currentColor, fit);
	                    if (d != null) {
	                        fit = d.intValue();
	                        fitRGB = pixelRGB;
	                    }
	                }
	                colorDictionary.put(pixel, new Integer(fitRGB));
	                newImage.setPixel(x, y, fitRGB);
	            }
	            Color actualColor = new Color(newImage.getPixel(x, y));
	            float redError = (currentColor.getRed() - actualColor.getRed()) / 16.0f;
	            float greenError = (currentColor.getGreen() - actualColor.getGreen()) / 16.0f;
	            float blueError = (currentColor.getBlue() - actualColor.getBlue()) / 16.0f;
	            if (x < width - 1) {
	                Color rightColor = new Color(newImage.getPixel(x + 1, y));
	                rightColor =
	                    new Color(
	                        Math.min(255, Math.max(0, rightColor.getRed() + (int) (redError * 7))),
	                        Math.min(255, Math.max(0, rightColor.getGreen() + (int) (greenError * 7))),
	                        Math.min(255, Math.max(0, rightColor.getBlue() + (int) (blueError * 7))));
	                newImage.setPixel(x + 1, y, rightColor.getRGB());
	            }
	            if (y < height - 1) {
	                if (x > 0) {
	                    Color leftDownColor = new Color(newImage.getPixel(x - 1, y + 1));
	                    leftDownColor =
	                        new Color(
	                            Math.min(255, Math.max(0, leftDownColor.getRed() + (int) (redError * 3))),
	                            Math.min(255, Math.max(0, leftDownColor.getGreen() + (int) (greenError * 3))),
	                            Math.min(255, Math.max(0, leftDownColor.getBlue() + (int) (blueError * 3))));
	                    newImage.setPixel(x - 1, y + 1, leftDownColor.getRGB());
	                }
	                Color centerDownColor = new Color(newImage.getPixel(x, y + 1));
	                centerDownColor =
	                    new Color(
	                        Math.min(255, Math.max(0, centerDownColor.getRed() + (int) (redError * 5))),
	                        Math.min(255, Math.max(0, centerDownColor.getGreen() + (int) (greenError * 5))),
	                        Math.min(255, Math.max(0, centerDownColor.getBlue() + (int) (blueError * 5))));
	                newImage.setPixel(x, y + 1, centerDownColor.getRGB());
	                if (x < newImage.width() - 1) {
	                    Color rightDownColor = new Color(newImage.getPixel(x + 1, y + 1));
	                    rightDownColor =
	                        new Color(
	                            Math.min(255, Math.max(0, rightDownColor.getRed() + (int) (redError))),
	                            Math.min(255, Math.max(0, rightDownColor.getGreen() + (int) (greenError))),
	                            Math.min(255, Math.max(0, rightDownColor.getBlue() + (int) (blueError))));
	                    newImage.setPixel(x + 1, y + 1, rightDownColor.getRGB());
	                }
	            }
	        }
	    }
	    return newImage;
	}
	/**
	 * Answer image representing my contents using the palette aPalette.
	 * Use NearistPatintImageRender to render the new image.
	 * 
	 * @return sra.smalltalk.StImage
	 * @param aPalette java.awt.image.IndexColorModel
	 */
	public StImage _convertToPalette_RenderedByNearistPaint(IndexColorModel aPalette) {
	    int width = this.width();
	    int height = this.height();
	    StImage newImage = new StImage(width, height);
	    Hashtable colorDictionary = new Hashtable();
	    for (int y = 0; y < height; y++) {
	        for (int x = 0; x < width; x++) {
	            Integer pixel = new Integer(this.getPixel(x, y));
	            if (colorDictionary.containsKey(pixel)) {
	                newImage.setPixel(x, y, ((Integer) colorDictionary.get(pixel)).intValue());
	            } else {
	                int fit = StColorValue._MaxDistanceSquared() + 1;
	                int fitRGB = 0xFF000000;
	                Color oldColor = new Color(pixel.intValue());
	                for (int j = 0; j < aPalette.getMapSize(); j++) {
	                    int pixelRGB = aPalette.getRGB(j);
	                    Color newColor = new Color(pixelRGB);
	                    Integer d =
	                        StColorValue._DistanceSquaredFrom_ifLessThan_(newColor, oldColor, fit);
	                    if (d != null) {
	                        fit = d.intValue();
	                        fitRGB = pixelRGB;
	                    }
	                }
	                colorDictionary.put(pixel, new Integer(fitRGB));
	                newImage.setPixel(x, y, fitRGB);
	            }
	        }
	    }
	    return newImage;
	}
    /**
     * display myself for new window.
     */
    public void _display() {
        StImage.Display_(this);
    }
	/**
	 * Flush the bit array.
	 */
	protected void _flushBits() {
	    bits = null;
	}
    /**
     * Answer an image from user.
     * @return sra.smalltalk.StImage
     */
    public static StImage _FromUser() {
        SystemInterface StPLInterface = SystemInterface.Current();
        int[] pixels, dimension = new int[2];
        synchronized (StPLInterface) {
            pixels = StPLInterface.utilImageFromUser(dimension);
        }
        if (pixels != null) {
            return new StImage(dimension[0], dimension[1], pixels);
        } else {
            return null;
        }
    }
    /**
     * Answer the pixel value as a 1-bit color image at the specified point in bits.
     *
     * @return int
     * @param x int
     * @param y int
     * @param bits int[]
     */
    protected int _getPixel1FromBits(int x, int y, byte[] bits) {
        int pixel = bits[y * rowByteSize + (x >> 3)] & 0xff;
        pixel = pixel >> (7 - (x & 7));
        pixel = pixel & 1;
        return this.image().getColorModel().getRGB(pixel);
    }
    /**
     * Answer the pixel value as a 24-bit color image at the specified point in bits.
     *
     * @return int
     * @param x int
     * @param y int
     * @param bits int[]
     */
    protected int _getPixel24FromBits(int x, int y, byte[] bits) {
        int byteIndex = y * rowByteSize + x + x + x;
        int r = bits[byteIndex + 0] & 0xff;
        int g = bits[byteIndex + 1] & 0xff;
        int b = bits[byteIndex + 2] & 0xff;
        return this.image().getColorModel().getRGB((new Color(r, g, b)).getRGB());
    }
    /**
     * Answer the pixel value as a 32-bit color image at the specified point in bits..
     *
     * @return int
     * @param x int
     * @param y int
     * @param bits byte[]
     */
    protected int _getPixel32FromBits(int x, int y, byte[] bits) {
        int byteIndex = y * rowByteSize + x + x + x + x;
        int a = 0xff - (bits[byteIndex + 0] & 0xff);
        int r = bits[byteIndex + 1] & 0xff;
        int g = bits[byteIndex + 2] & 0xff;
        int b = bits[byteIndex + 3] & 0xff;
        return this.image().getColorModel().getRGB((new Color(r, g, b, a)).getRGB());
    }
    /**
     * Answer the pixel value at the specified point in bits.
     *
     * @return int
     * @param x int
     * @param y int
     * @param bits byte[]
     */
    protected int _getPixelFromBits(int x, int y, byte[] bits) {
        int bitsPerPixel = this.bitsPerPixel();
        if (bitsPerPixel == 32) {
            return this._getPixel32FromBits(x, y, bits);
        } else if (bitsPerPixel == 24) {
            return this._getPixel24FromBits(x, y, bits);
        } else if (bitsPerPixel == 1) {
            return this._getPixel1FromBits(x, y, bits);
        }
        throw new SmalltalkException(
            "can't handle an " + bitsPerPixel + "-color image");
    }
    /**
     * Answer an image of specified area of screen.
     * @return sra.smalltalk.StImage
     * @param rect java.awt.Rectangle
     */
    public static StImage _OfArea(Rectangle rect) {
        SystemInterface StPLInterface = SystemInterface.Current();
        int[] pixels;
        synchronized (StPLInterface) {
            pixels = StPLInterface.utilImageOfArea(rect.x, rect.y, rect.width, rect.height);
        }
        if (pixels != null) {
            return new StImage(rect.width, rect.height, pixels);
        } else {
            return null;
        }
    }
    /**
     * Answer an image of whole screen.
     * @return sra.smalltalk.StImage
     */
    public static StImage _OfWholeScreen() {
        SystemInterface StPLInterface = SystemInterface.Current();
        int[] pixels, dimension = new int[2];
        synchronized (StPLInterface) {
            pixels = StPLInterface.utilImageOfWholeScreen(dimension);
        }
        if (pixels != null) {
            return new StImage(dimension[0], dimension[1], pixels);
        } else {
            return null;
        }
    }
	/**
	 * Set the pixel value as a 1-bit color image at the specified point in bits.
	 *
	 * @return int
	 * @param x int
	 * @param y int
	 * @param bits byte[]
	 */
	protected int _setPixel1IntoBits(int x, int y, byte[] bits) {
	    ColorModel colorModel = this.image().getColorModel();
	    int pixelValue = colorModel.getRGB(this.getPixel(x, y) & 0xff);
	    int byteIndex = y * rowByteSize + (x >> 3);
	    int whichBit = 7 - (x & 7);
	    int value =
	        (pixelValue == 0)
	            ? bits[byteIndex] & (-1 - (1 << whichBit))
	            : bits[byteIndex] | (1 << whichBit);
	    bits[byteIndex] = (byte) value;
	    return pixelValue;
	}
    /**
     * Set the pixel value as a 24-bit color image at the specified point in bits.
     *
     * @return int
     * @param x int
     * @param y int
     * @param bits byte[]
     */
    protected int _setPixel24IntoBits(int x, int y, byte[] bits) {
        ColorModel colorModel = this.image().getColorModel();
        int pixelValue = this.getPixel(x, y);
        int byteIndex = y * rowByteSize + x + x + x;
        bits[byteIndex + 0] = (byte) colorModel.getRed(pixelValue);
        bits[byteIndex + 1] = (byte) colorModel.getGreen(pixelValue);
        bits[byteIndex + 2] = (byte) colorModel.getBlue(pixelValue);
        return pixelValue;
    }
    /**
     * Set the pixel value as a 32-bit color image at the specified point in bits.
     *
     * @return int
     * @param x int
     * @param y int
     * @param bits byte[]
     */
    protected int _setPixel32IntoBits(int x, int y, byte[] bits) {
        ColorModel colorModel = this.image().getColorModel();
        int pixelValue = this.getPixel(x, y);
        int byteIndex = y * rowByteSize + x + x + x + x;
        bits[byteIndex + 0] = (byte) (0xff - colorModel.getAlpha(pixelValue));
        bits[byteIndex + 1] = (byte) colorModel.getRed(pixelValue);
        bits[byteIndex + 2] = (byte) colorModel.getGreen(pixelValue);
        bits[byteIndex + 3] = (byte) colorModel.getBlue(pixelValue);
        return pixelValue;
    }
    /**
     * Set the pixel value at the specified point in bits.
     *
     * @return int
     * @param x int
     * @param y int
     * @param bits byte[]
     */
    protected int _setPixelIntoBits(int x, int y, byte[] bits) {
        int bitsPerPixel = this.bitsPerPixel();
        if (bitsPerPixel == 32) {
            return this._setPixel32IntoBits(x, y, bits);
        } else if (bitsPerPixel == 24) {
            return this._setPixel24IntoBits(x, y, bits);
        } else if (bitsPerPixel == 1) {
            return this._setPixel1IntoBits(x, y, bits);
        }
        throw new SmalltalkException(
            "can't handle an " + bitsPerPixel + "-color image");
    }
    /**
     * Answer the pixel at the specified zero-based 2D index.
     *
     * @deprecated use getPixel(int x, int y) instead.
     * @return int
     * @param point java.awt.Point
     */
    public int atPoint_(Point point) {
        return this.getPixel(point.x, point.y);
    }
    /**
     * Set a pixel value at the specified zero-based 2D index.
     *
     * @deprecated use setPixel(int x, int y, int pixel) instead.
     * @param point java.awt.Point
     * @param pixelValue int
     */
    public void atPoint_put_(Point point, int pixelValue) {
        this.setPixel(point.x, point.y, pixelValue);
    }
    /**
     * Answer the pixel at location (x, y).
     *
     * @deprecated use getPixel(int x, int y) instead.
     * @return int
     * @param x int
     * @param y int
     */
    public int atX_y_(int x, int y) {
        return this.getPixel(x, y);
    }
    /**
     * Set the pixel at location (x, y).
     *
     * @deprecated use setPixel(int x, int y, int pixel) instead.
     * @param x int
     * @param y int
     * @param pixelValue int
     */
    public void atX_y_put_(int x, int y, int pixelValue) {
        this.setPixel(x, y, pixelValue);
    }
	/**
	 * Answer the bitmap contents of this image.
	 * 
	 * @return byte[]
	 */
	protected byte[] bits() {
	    if (bits == null) {
	        int width = this.width();
	        int height = this.height();
	        bits = new byte[rowByteSize * height];
	        for (int y = 0; y < height; y++) {
	            for (int x = 0; x < width; x++) {
	                this._setPixelIntoBits(x, y, bits);
	            }
	        }
	    }
	    return bits;
	}
    /**
     * Answer the number of the bits for one pixel.
     *
     * @return int
     */
    public int bitsPerPixel() {
        return this.image().getColorModel().getPixelSize();
    }
    /**
     * Answer the bounds of this image.
     * 
     * @return java.awt.Rectangle
     */
    public Rectangle bounds() {
        return new Rectangle(0, 0, this.width(), this.height());
    }
	/**
	 * Create a copy of the StImage.
	 *
	 * @return java.lang.Object
	 */
	public Object clone() {
	    return new StImage(
	        this.width(),
	        this.height(),
	        this.getPixels(),
	        this.image().getColorModel());
	}
    /**
     * Answer image representing my contents using the palette aPalette.
     * 
     * @return sra.smalltalk.StImage
     * @param aPalette java.awt.image.ColorModel
     */
    public StImage convertToPalette_(IndexColorModel aPalette) {
        return this._convertToPalette_RenderedByNearistPaint(aPalette);
    }
	/**
	 * Copy into the destination rectangle destRectangle in the receiver 
	 * those bits in the sourceImage starting at position sourcePt,
	 * according the combination rule rule.
	 *
	 * @return StImage
	 * @param destRectangle java.awt.Rectangle
	 * @param sourcePt java.awt.Point
	 * @param sourceImage sra.smalltalk.StImage
	 * @param combinationRule int
	 */
	public StImage copy_from_in_rule_(
	    Rectangle destRectangle,
	    Point sourcePt,
	    StImage sourceImage,
	    int combinationRule) {

	    int width = destRectangle.width;
	    int height = destRectangle.height;
	    int destOriginX = destRectangle.x;
	    int destOriginY = destRectangle.y;
	    int destCornerX = destRectangle.x + width - 1;
	    int destCornerY = destRectangle.y + height - 1;
	    int sourceOriginX = sourcePt.x;
	    int sourceOriginY = sourcePt.y;
	    int sourceCornerX = sourcePt.x + width - 1;
	    int sourceCornerY = sourcePt.y + height - 1;
	    if (this.bounds().contains(destRectangle.getLocation()) == false) {
	        return null;
	    }
	    if (this.bounds().contains(destCornerX, destCornerY) == false) {
	        return null;
	    }
	    if (sourceImage.bounds().contains(sourcePt) == false) {
	        return null;
	    }
	    if (sourceImage.bounds().contains(sourceCornerX, sourceCornerY) == false) {
	        return null;
	    }
	    for (int y = 0; y < height; y++) {
	        for (int x = 0; x < width; x++) {
	            Point sourcePoint = new Point(x + sourceOriginX, y + sourceOriginY);
	            Point destinationPoint = new Point(x + destOriginX, y + destOriginY);
	            int newPaint =
	                this._combinationedColorFrom_to_in_(
	                    sourceImage.atPoint_(sourcePoint),
	                    this.atPoint_(destinationPoint),
	                    combinationRule);
	            this.atPoint_put_(destinationPoint, newPaint);
	        }
	    }
	    return this;
	}
    /**
     * Answer a new Image with the same extent, depth as the this image.
     * 
     * @return sra.smalltalk.StImage
     * @param newExtent java.awt.Point
     */
    public StImage copyEmpty_(Point newExtent) {
        return new StImage(newExtent.x, newExtent.y);
    }
    /**
     * Display image for new window
     *
     * @param anImage sra.smalltalk.StImage
     */
    public static void Display_(StImage anImage) {
        Frame aFrame = new Frame();
        aFrame.setTitle("Display");
        aFrame.setLayout(new BorderLayout());
        aFrame.addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent e) {
                e.getWindow().dispose();
            }
        });
        final StImage _image = anImage;
        Canvas newCanvas = new Canvas() {
            public void paint(Graphics g) {
                _image.displayOn_(g);
                return;
            }
        };
        newCanvas.setBounds(0, 0, anImage.width(), anImage.height());
        aFrame.add(newCanvas);
        aFrame.pack();
        StApplicationModel._ShowAtMousePoint(aFrame);
    }
    /**
     * Update the view.
     *
     * @param aGraphics java.awt.Graphics
     */
    public void displayOn_(Graphics aGraphics) {
        this.displayOn_at_(aGraphics, new Point(0, 0));
    }
    /**
     * Update the view.
     *
     * @param aGraphics java.awt.Graphics
     * @param aPoint java.awt.Point
     */
    public void displayOn_at_(Graphics aGraphics, Point aPoint) {
        Graphics graphics = aGraphics.create();
        try {
            Canvas medium = new Canvas();
            MediaTracker mt = new MediaTracker(medium);
            mt.addImage(this.image(), 0);
            graphics.drawImage(this.image(), aPoint.x, aPoint.y, medium);
        } finally {
            graphics.dispose();
        }
    }
    /**
     * Answer extent of this image.
     *
     * @return java.awt.Point
     */
    public Point extent() {
        return new Point(this.width(), this.height());
    }
    /**
     * Answer the pixel value at the specified point.
     *
     * @return int
     * @param x int
     * @param y int
     */
    public int getPixel(int x, int y) {
        return this.image().getRGB(x, y);
    }
	/**
	 * Answer all of the pixel values.
	 *
	 * @return int[]
	 */
	public int[] getPixels() {
	    int width = this.width();
	    int height = this.height();
	    return this.image().getRGB(0, 0, width, height, null, 0, width);
	}
    /**
     * Answer the height of this image.
     *
     * @return int
     */
    public int height() {
        return this.image().getHeight();
    }
    /**
     * Answer the image object.
     *
     * @return java.awt.image.BufferedImage
     */
    public BufferedImage image() {
        return image;
    }
	/**
	 * Restores this object from a object stream (i.e., deserializes it).
	 *
	 * @param stream java.io.ObjectInputStream
	 * @exception java.io.IOException
	 * @exception java.lang.ClassNotFoundException
	 */
	private void readObject(ObjectInputStream stream) throws IOException, ClassNotFoundException {
		stream.defaultReadObject();

		// read width, height and pixels.
		int width = stream.readInt();
		int height = stream.readInt();
		int[] pixels = (int[]) stream.readObject();

		// read ColorModel.
		String colorModelName = stream.readUTF();
		ColorModel colorModel = null;
		if (colorModelName.equals("DirectColorModel")) {
			int pixelSize = stream.readInt();
			int redMask = stream.readInt();
			int greenMask = stream.readInt();
			int blueMask = stream.readInt();
			int alphaMask = stream.readInt();
			//
			colorModel = new DirectColorModel(pixelSize, redMask, greenMask, blueMask, alphaMask);
		} else if (colorModelName.equals("IndexColorModel")) {
			int pixelSize = stream.readInt();
			int mapSize = stream.readInt();
			byte[] r = new byte[mapSize];
			byte[] g = new byte[mapSize];
			byte[] b = new byte[mapSize];
			byte[] a = new byte[mapSize];
			for (int i = 0; i < mapSize; i++) {
				r[i] = (byte) stream.read();
			}
			for (int i = 0; i < mapSize; i++) {
				g[i] = (byte) stream.read();
			}
			for (int i = 0; i < mapSize; i++) {
				b[i] = (byte) stream.read();
			}
			for (int i = 0; i < mapSize; i++) {
				a[i] = (byte) stream.read();
			}
			//
			colorModel = new IndexColorModel(pixelSize, mapSize, r, g, b, a);
		} else {
			throw new IOException("The color model [" + colorModel.getClass() + "] is not supported yet.");
		}

		// create image.
		image = new BufferedImage(colorModel, colorModel.createCompatibleWritableRaster(width, height), false, null);
		rowByteSize = (width * this.bitsPerPixel() + 31) >> 5 << 2;
		this.setPixels(pixels);
	}
    /**
     * Answer resized image with (x, y).
     * 
     * @return java.awt.Point
     * @param scaleX double
     * @param scaleY double
     */
    public Point scaledExtent_(double scaleX, double scaleY) {
        Double x = new Double(this.width() * 1.0 / scaleX);
        Double y = new Double(this.height() * 1.0 / scaleY);
        return new Point(x.intValue(), y.intValue());
    }
    /**
     * Set the pixel value to the specified point.
     *
     * @param x int
     * @param y int
     * @param pixel int
     */
    public void setPixel(int x, int y, int pixel) {
        this.image().setRGB(x, y, pixel);
        this._flushBits();
    }
	/**
	 * Set the specified pixel values.
	 *
	 * @param pixels int[]
	 */
	public void setPixels(int[] pixels) {
	    int width = this.width();
	    int height = this.height();

	    if (pixels == null) {
	        pixels = new int[width * height];
	    } else if (width * height > pixels.length) {
	        throw SmalltalkException.Error("Pixel array has incorrect size.");
	    }

	    this.image().setRGB(0, 0, width, height, pixels, 0, width);
	    this._flushBits();
	}
    /**
     * Print my storeable string representation on the writer.
     * 
     * @param aWriter java.io.Writer
     * @exception java.io.IOException The exception description.
     */
    public void storeOn_(Writer aWriter) throws IOException {
        aWriter.write("(Image extent: ");
        aWriter.write(String.valueOf(this.width()));
        aWriter.write('@');
        aWriter.write(String.valueOf(this.height()));
        aWriter.write(" depth: ");
        aWriter.write(String.valueOf(this.bitsPerPixel()));
        aWriter.write(" bitsPerPixel: ");
        aWriter.write(String.valueOf(this.bitsPerPixel()));
        aWriter.write(" palette: ");
        this.storePaletteOn_(aWriter);
        aWriter.write(" usingBits: ");
        (new StByteArray(this.bits())).storeOn_(aWriter);
        aWriter.write(')');
    }
	/**
	 * Print the storeable string representation of my palette on the writer.
	 *
	 * @param aWriter java.io.Writer
	 * @exception java.io.IOException The exception description.
	 */
	protected void storePaletteOn_(Writer aWriter) throws IOException {
	    ColorModel colorModel = this.image().getColorModel();
	    if (colorModel instanceof DirectColorModel) {
	        aWriter.write("(FixedPalette redShift: ");
	        aWriter.write(
	            String.valueOf(
	                (int) (Math.log(((DirectColorModel) colorModel).getRedMask() / 255)
	                    / Math.log(2))));
	        aWriter.write(" redMask: 255 greenShift: ");
	        aWriter.write(
	            String.valueOf(
	                (int) (Math.log(((DirectColorModel) colorModel).getGreenMask() / 255)
	                    / Math.log(2))));
	        aWriter.write(" greenMask: 255 blueShift: ");
	        aWriter.write(
	            String.valueOf(
	                (int) (Math.log(((DirectColorModel) colorModel).getBlueMask() / 255)
	                    / Math.log(2))));
	        aWriter.write(" blueMask: 255)");
	    } else if (colorModel instanceof IndexColorModel) {
	        IndexColorModel indexColorModel = (IndexColorModel) colorModel;
	        int map_size = indexColorModel.getMapSize();
	        aWriter.write("(MappedPalette withColors: ");
	        aWriter.write("((Array new: " + map_size + ") ");
	        for (int i = 0; i < map_size; i++) {
	            aWriter.write("at: " + (i + 1) + " put: ");
	            aWriter.write(StColorValue._StoreStringOf(new Color(indexColorModel.getRGB(i))));
	            aWriter.write("; ");
	        }
	        aWriter.write(" yourself))");
	    } else {
	        throw new SmalltalkException(
	            "The color model [" + colorModel.getClass() + "] is not supported yet.");
	    }
	}
    /**
     * Tile the destination rectangle destRectangle in the receiver
     * with copies of the sourceImage, according to the combination rule rule.
     * The sourcePt in the sourceImage is aligned with 0@0 in the receiver.
     * 
     * @return StImage
     * @param destRectangle java.awt.Rectangle
     * @param sourcePt java.awt.Point
     * @param sourceImage sra.smalltalk.StImage
     * @param combinationRule int
     */
    public StImage tile_from_in_rule_(
        Rectangle destRectangle,
        Point sourcePt,
        StImage sourceImage,
        int combinationRule) {
        int width = destRectangle.width;
        int height = destRectangle.height;
        int sourceWidth = sourceImage.width();
        int sourceHeight = sourceImage.height();
        int destOriginX = destRectangle.x;
        int destOriginY = destRectangle.y;
        int destCornerX = destRectangle.x + width - 1;
        int destCornerY = destRectangle.y + height - 1;
        int sourceOriginX = sourcePt.x;
        int sourceOriginY = sourcePt.y;
        int sourceCornerX = sourcePt.x + width - 1;
        int sourceCornerY = sourcePt.y + height - 1;
        if (this.bounds().contains(destRectangle.getLocation()) == false) {
            return null;
        }
        if (this.bounds().contains(destCornerX, destCornerY) == false) {
            return null;
        }
        for (int y = 0; y < height; y++) {
            for (int x = 0; x < width; x++) {
                Point sourcePoint =
                    new Point(
                        (x + sourceOriginX) % sourceWidth,
                        (y + sourceOriginY) % sourceHeight);
                Point destinationPoint = new Point(x + destOriginX, y + destOriginY);
                int newPaint =
                    this._combinationedColorFrom_to_in_(
                        sourceImage.atPoint_(sourcePoint),
                        this.atPoint_(destinationPoint),
                        combinationRule);
                this.atPoint_put_(destinationPoint, newPaint);
            }
        }
        return this;
    }
	/**
	 * Answer the value at aPoint according to this image interpretation.
	 * 
	 * @return java.awt.Color
	 * @param point java.awt.Point
	 */
	public Color valueAtPoint_(Point point) {
	    ColorModel colorModel = this.image().getColorModel();
	    int pixel = this.getPixel(point.x, point.y);
	    return new Color(
	        colorModel.getRed(pixel),
	        colorModel.getGreen(pixel),
	        colorModel.getBlue(pixel),
	        colorModel.getAlpha(pixel));
	}
	/**
	 * Set the value at aPoint according to this image interpretation.
	 * 
	 * @param point java.awt.Point
	 * @param color java.awt.Color
	 */
	public void valueAtPoint_put_(Point point, Color color) {
	    this.setPixel(point.x, point.y, color.getRGB());
	}
    /**
     * Answer the width of this image.
     *
     * @return int
     */
    public int width() {
        return this.image().getWidth();
    }
	/**
	 * Writes this object out to a stream (i.e., serializes it).
	 *
	 * @param stream java.io.ObjectOutputStream
	 * @exception java.io.IOException
	 * @exception java.lang.ClassNotFoundException
	 */
	private void writeObject(ObjectOutputStream stream) throws IOException, ClassNotFoundException {
		stream.defaultWriteObject();

		// write width, height and pixels.
		stream.writeInt(this.width());
		stream.writeInt(this.height());
		stream.writeObject(this.getPixels());

		// write ColorModel.
		ColorModel colorModel = this.image().getColorModel();
		if (colorModel instanceof DirectColorModel) {
			DirectColorModel directColorModel = (DirectColorModel) colorModel;
			stream.writeUTF("DirectColorModel");
			stream.writeInt(directColorModel.getPixelSize());
			stream.writeInt(directColorModel.getRedMask());
			stream.writeInt(directColorModel.getGreenMask());
			stream.writeInt(directColorModel.getBlueMask());
			stream.writeInt(directColorModel.getAlphaMask());
		} else if (colorModel instanceof IndexColorModel) {
			IndexColorModel indexColorModel = (IndexColorModel) colorModel;
			stream.writeUTF("IndexColorModel");
			stream.writeInt(indexColorModel.getPixelSize());
			stream.writeInt(indexColorModel.getMapSize());
			byte[] bytes = new byte[indexColorModel.getMapSize()];
			indexColorModel.getReds(bytes);
			stream.write(bytes, 0, bytes.length);
			indexColorModel.getGreens(bytes);
			stream.write(bytes, 0, bytes.length);
			indexColorModel.getBlues(bytes);
			stream.write(bytes, 0, bytes.length);
			indexColorModel.getAlphas(bytes);
			stream.write(bytes, 0, bytes.length);
		} else {
			throw new IOException("The color model [" + colorModel.getClass() + "] is not supported yet.");
		}
	}
}
